<?php

/**
 * Copyright 2017 DataStax, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

namespace Cassandra;

/**
 * Common interface implemented by all Cassandra value types.
 *
 * @see Bigint
 * @see Smallint
 * @see Tinyint
 * @see Blob
 * @see Collection
 * @see Float
 * @see Inet
 * @see Map
 * @see Set
 * @see Timestamp
 * @see Timeuuid
 * @see Uuid
 * @see Varint
 * @see Date
 * @see Time
 *
 * @see Numeric
 * @see UuidInterface
 */
interface Value {

    /**
     * The type of represented by the value.
     *
     * @return Type
     */
    public function type();

}
